@@ -8,7 +8,7 @@
 
 #include <asm/types.h>          /* For __uXX types */
 
-#define IP_VS_VERSION_CODE            0x000905
+#define IP_VS_VERSION_CODE            0x000906
 #define NVERSION(version)                       \
 	(version >> 16) & 0xFF,                 \
 	(version >> 8) & 0xFF,                  \
@@ -64,7 +64,8 @@
 #define IP_VS_SO_SET_STOPDAEMON (IP_VS_BASE_CTL+12)
 #define IP_VS_SO_SET_RESTORE    (IP_VS_BASE_CTL+13)
 #define IP_VS_SO_SET_SAVE       (IP_VS_BASE_CTL+14)
-#define IP_VS_SO_SET_MAX	IP_VS_SO_SET_SAVE
+#define IP_VS_SO_SET_ZERO	(IP_VS_BASE_CTL+15)
+#define IP_VS_SO_SET_MAX	IP_VS_SO_SET_ZERO
 
 #define IP_VS_SO_GET_VERSION	IP_VS_BASE_CTL
 #define IP_VS_SO_GET_INFO	(IP_VS_BASE_CTL+1)
@@ -348,6 +349,10 @@ struct ip_vs_timeout_table {
 	int timeout[IP_VS_S_LAST+1];
 };
 
+
+/*
+ *	Transport protocol header
+ */
 union ip_vs_tphdr {
 	unsigned char *raw;
 	struct udphdr *uh;
@@ -545,9 +550,6 @@ struct ip_vs_app
  *      IPVS core functions
  *      (from ip_vs_core.c)
  */
-extern struct ip_vs_conn * ip_vs_schedule(struct ip_vs_service *svc,
-					  struct iphdr *iph);
-extern int ip_vs_proto_doff(unsigned proto, char *th, unsigned size);
 extern const char *ip_vs_proto_name(unsigned proto);
 extern unsigned int check_for_ip_vs_out(struct sk_buff **skb_p,
 					int (*okfn)(struct sk_buff *));
@@ -663,9 +665,8 @@ extern struct ip_vs_app * ip_vs_bind_app(struct ip_vs_conn *cp);
 extern int ip_vs_unbind_app(struct ip_vs_conn *cp);
 extern int ip_vs_app_pkt_out(struct ip_vs_conn *, struct sk_buff **skb_p);
 extern int ip_vs_app_pkt_in(struct ip_vs_conn *, struct sk_buff **skb_p);
-extern struct sk_buff * ip_vs_skb_replace (struct sk_buff *skb, int pri,
-					   char *o_buf, int o_len,
-					   char *n_buf, int n_len);
+extern int ip_vs_skb_replace(struct sk_buff *skb, int pri,
+			     char *o_buf, int o_len, char *n_buf, int n_len);
 extern int ip_vs_app_init(void);
 extern void ip_vs_app_cleanup(void);
 
@@ -676,12 +677,11 @@ extern void ip_vs_app_cleanup(void);
  */
 extern int register_ip_vs_scheduler(struct ip_vs_scheduler *scheduler);
 extern int unregister_ip_vs_scheduler(struct ip_vs_scheduler *scheduler);
-extern int ip_vs_register_scheduler_module(struct ip_vs_scheduler *scheduler);
-extern void ip_vs_unregister_scheduler_module(struct ip_vs_scheduler *scheduler);
 extern int ip_vs_bind_scheduler(struct ip_vs_service *svc,
 				struct ip_vs_scheduler *scheduler);
 extern int ip_vs_unbind_scheduler(struct ip_vs_service *svc);
-extern struct ip_vs_scheduler *ip_vs_lookup_scheduler(const char *sched_name);
+extern struct ip_vs_scheduler *ip_vs_scheduler_get(const char *sched_name);
+extern void ip_vs_scheduler_put(struct ip_vs_scheduler *scheduler);
 
 
 /*
@@ -717,8 +717,8 @@ extern void ip_vs_control_cleanup(void);
  *      (from ip_vs_sync.c)
  */
 extern int ip_vs_sync_state;
-extern char mcast_ifn[IP_VS_IFNAME_MAXLEN];
-extern int start_sync_thread(struct ip_vs_rule_user *u);
+extern char ip_vs_mcast_ifn[IP_VS_IFNAME_MAXLEN];
+extern int start_sync_thread(int state, char *mcast_ifn);
 extern int stop_sync_thread(void);
 extern void ip_vs_sync_conn(struct ip_vs_conn *cp);
 
@@ -728,6 +728,7 @@ extern void ip_vs_sync_conn(struct ip_vs_conn *cp);
  */
 int ip_vs_new_estimator(struct ip_vs_stats *stats);
 void ip_vs_kill_estimator(struct ip_vs_stats *stats);
+void ip_vs_zero_estimator(struct ip_vs_stats *stats);
 
 
 /*
@@ -786,17 +787,29 @@ extern __inline__ char ip_vs_fwd_tag(struct ip_vs_conn *cp)
 
 
 /*
- *	TCP/UDP header check (against IP payload size)
+ *	transport layer header checking
  */
-extern inline int ip_vs_header_check(unsigned proto, char *th, unsigned size)
+extern inline int ip_vs_header_check(struct sk_buff *skb, int proto, int ihl)
 {
-	if (proto == IPPROTO_TCP) {
-		unsigned short hlen = ((struct tcphdr*)th)->doff << 2;
-		if (hlen < sizeof(struct tcphdr) || hlen > size)
-			return -1;
-	} else if (proto == IPPROTO_UDP && size < sizeof(struct udphdr))
+	int len;
+
+	switch (proto) {
+	case IPPROTO_TCP:
+		len = ihl + sizeof(struct tcphdr);
+		/* we don't care about TCP options */
+		break;
+	case IPPROTO_UDP:
+		len = ihl + sizeof(struct udphdr);
+		break;
+	default:
+		len = 0;
+	}
+
+	/* guarantee protocol header available in skb data area */
+	if (!pskb_may_pull(skb, len))
 		return -1;
-	return 0;
+	else
+		return 0;
 }
 
 
@@ -898,27 +911,28 @@ static inline void ip_vs_fast_check_update(union ip_vs_tphdr *h,
 		*checkp = 0xFFFF;
 }
 
-/*
- *	Own skb_cow() beast, tweaked for rewriting commonly
- *	used pointers in the code
- */
-static inline struct sk_buff *vs_skb_cow(struct sk_buff **skb_p,
-				struct iphdr **iph_p, unsigned char **t_p)
+static inline struct sk_buff *
+ip_vs_skb_cow(struct sk_buff **skb_p, struct iphdr **iph_p,
+	      unsigned char **t_p)
 {
-	struct sk_buff *skb=(*skb_p);
+	struct sk_buff *skb = *skb_p;
+
 	if (skb_cloned(skb)) {
-		skb = skb_copy(skb, GFP_ATOMIC);
-		if (skb) {
-			/*
-			 *	skb changed, update other pointers
-			 */
-			struct iphdr *iph = skb->nh.iph;
-			kfree_skb(*skb_p);
-			*skb_p = skb;
-			*iph_p = iph;
-			*t_p = (char*) iph + iph->ihl * 4;
-		}
+		struct iphdr *iph;
+
+		skb = pskb_copy(skb, GFP_ATOMIC);
+		if (!skb)
+			return NULL;
+		/*
+		 *	skb changed, update other pointers
+		 */
+		kfree_skb(*skb_p);
+		*skb_p = skb;
+		iph = skb->nh.iph;
+		*iph_p = iph;
+		*t_p = (char*) iph + iph->ihl * 4;
 	}
+
 	return skb;
 }
 
