@@ -1,7 +1,7 @@
 /*
  * IPVS         Application module
  *
- * Version:     $Id: ip_vs_app.c,v 1.12 2001/05/09 15:01:14 wensong Exp $
+ * Version:     $Id: ip_vs_app.c,v 1.13 2001/11/04 08:58:43 wensong Exp $
  *
  * Authors:     Wensong Zhang <wensong@linuxvirtualserver.org>
  *
@@ -97,17 +97,6 @@ int unregister_ip_vs_app(struct ip_vs_app *vapp)
 		return -EINVAL;
 	}
 
-#if 0000
-	/*
-	 * only allow unregistration if it has no referrer.
-	 */
-	if (atomic_read(&vapp->refcnt) > 0)  {
-		IP_VS_ERR("unregister_ip_vs_app(): has %d referrers. failed\n",
-			  atomic_read(&vapp->refcnt));
-		return -EINVAL;
-	}
-#endif
-
 	write_lock_bh(&__ip_vs_app_lock);
 	list_del(&vapp->n_list);
 	write_unlock_bh(&__ip_vs_app_lock);
@@ -187,7 +176,6 @@ struct ip_vs_app * ip_vs_bind_app(struct ip_vs_conn *cp)
 	vapp = ip_vs_app_get(cp->protocol, cp->vport);
 
 	if (vapp != NULL) {
-		/*  atomic_inc(&vapp->refcnt); */
 		cp->app = vapp;
 
 		if (vapp->init_conn)
@@ -211,7 +199,6 @@ int ip_vs_unbind_app(struct ip_vs_conn *cp)
 		if (vapp->done_conn)
 			vapp->done_conn(vapp, cp);
 		cp->app = NULL;
-		/*  atomic_dec(&vapp->refcnt); */
 		if (vapp->module)
 			__MOD_DEC_USE_COUNT(vapp->module);
 	}
@@ -312,7 +299,7 @@ int ip_vs_app_pkt_out(struct ip_vs_conn *cp, struct sk_buff **skb_p)
 	 *	check if application module is bound to
 	 *	this ip_vs_conn.
 	 */
-	if ( (vapp = cp->app) == NULL)
+	if ((vapp = cp->app) == NULL)
 		return 0;
 
 	iph = (*skb_p)->nh.iph;
@@ -336,7 +323,7 @@ int ip_vs_app_pkt_out(struct ip_vs_conn *cp, struct sk_buff **skb_p)
 	/*
 	 *	Call private output hook function
 	 */
-	if ( vapp->pkt_out == NULL )
+	if (vapp->pkt_out == NULL)
 		return 0;
 
 	diff = vapp->pkt_out(vapp, cp, skb_p);
@@ -369,7 +356,7 @@ int ip_vs_app_pkt_in(struct ip_vs_conn *cp, struct sk_buff **skb_p)
 	 *	check if application module is bound to
 	 *	this ip_vs_conn.
 	 */
-	if ( (vapp = cp->app) == NULL)
+	if ((vapp = cp->app) == NULL)
 		return 0;
 
 	iph = (*skb_p)->nh.iph;
@@ -393,7 +380,7 @@ int ip_vs_app_pkt_in(struct ip_vs_conn *cp, struct sk_buff **skb_p)
 	/*
 	 *	Call private input hook function
 	 */
-	if ( vapp->pkt_in == NULL )
+	if (vapp->pkt_in == NULL)
 		return 0;
 
 	diff = vapp->pkt_in(vapp, cp, skb_p);
@@ -460,115 +447,46 @@ static int ip_vs_app_getinfo(char *buffer, char **start, off_t offset,
 
 
 /*
- *	Replace a segment (of skb->data) with a new one.
- *	FIXME: Should re-use same skb if space available, this could
- *	       be done if n_len < o_len, unless some extra space
- *	       were already allocated at driver level :P .
+ *	Replace a segment of data with a new segment
  */
-static struct sk_buff *skb_replace(struct sk_buff *skb, int pri, char *o_buf,
-				   int o_len, char *n_buf, int n_len)
-{
-	int maxsize, diff, o_offset;
-	struct sk_buff *n_skb;
-	int offset;
-
-	maxsize = skb->truesize;
-	diff = n_len - o_len;
-	o_offset = o_buf - (char*) skb->data;
-
-	if (maxsize <= n_len) {
-		if (diff != 0) {
-			memcpy(skb->data + o_offset + n_len,o_buf + o_len,
-			       skb->len - (o_offset + o_len));
-		}
-
-		memcpy(skb->data + o_offset, n_buf, n_len);
-
-		n_skb    = skb;
-		skb->len = n_len;
-		skb->end = skb->head+n_len;
-	} else {
-		/*
-		 *	Sizes differ, make a copy.
-		 *
-		 *	FIXME: move this to core/sbuff.c:skb_grow()
-		 */
-		n_skb = alloc_skb(MAX_HEADER + skb->len + diff, pri);
-		if (n_skb == NULL) {
-			IP_VS_ERR("skb_replace(): no room left (from %p)\n",
-				  __builtin_return_address(0));
-			return skb;
-
-		}
-		skb_reserve(n_skb, MAX_HEADER);
-		skb_put(n_skb, skb->len + diff);
-
-		/*
-		 *	Copy as much data from the old skb as possible. Even
-		 *	though we're only forwarding packets, we need stuff
-		 *	like skb->protocol (PPP driver wants it).
-		 */
-		offset = n_skb->data - skb->data;
-		n_skb->nh.raw = skb->nh.raw + offset;
-		n_skb->h.raw = skb->h.raw + offset;
-		n_skb->dev = skb->dev;
-		n_skb->mac.raw = skb->mac.raw + offset;
-		n_skb->pkt_type = skb->pkt_type;
-		n_skb->protocol = skb->protocol;
-		n_skb->ip_summed = skb->ip_summed;
-		n_skb->dst = dst_clone(skb->dst);
-
-		/*
-		 * Copy pkt in new buffer
-		 */
-		memcpy(n_skb->data, skb->data, o_offset);
-		memcpy(n_skb->data + o_offset, n_buf, n_len);
-		memcpy(n_skb->data + o_offset + n_len, o_buf + o_len,
-		       skb->len - (o_offset + o_len) );
-
-		/*
-		 * Problem, how to replace the new skb with old one,
-		 * preferably inplace
-		 */
-		kfree_skb(skb);
-	}
-
-	return n_skb;
-}
-
-
-/*
- *	calls skb_replace() and update ip header if new skb was allocated
- */
-struct sk_buff *ip_vs_skb_replace(struct sk_buff *skb, int pri, char *o_buf,
-				  int o_len, char *n_buf, int n_len)
+int ip_vs_skb_replace(struct sk_buff *skb, int pri,
+		      char *o_buf, int o_len, char *n_buf, int n_len)
 {
+	struct iphdr *iph;
 	int diff;
-	struct sk_buff *n_skb;
-	unsigned skb_len;
+	int o_offset;
+	int o_left;
 
 	EnterFunction(9);
 
 	diff = n_len - o_len;
-	n_skb = skb_replace(skb, pri, o_buf, o_len, n_buf, n_len);
-	skb_len = skb->len;
+	o_offset = o_buf - (char *)skb->data;
+	/* The length of left data after o_buf+o_len in the skb data */
+	o_left = skb->len - (o_offset + o_len);
 
-	if (diff)
-	{
-		struct iphdr *iph;
-		IP_VS_DBG(9, "pkt resized for %d bytes (len=%d)\n",
-			  diff, skb->len);
-		/*
-		 *	update ip header
-		 */
-		iph = n_skb->nh.iph;
-		iph->tot_len = htons(skb_len + diff);
-		iph->check = 0;
-		iph->check = ip_fast_csum((unsigned char *)iph, iph->ihl);
+	if (diff <= 0) {
+		memmove(o_buf + n_len, o_buf + o_len, o_left);
+		memcpy(o_buf, n_buf, n_len);
+		skb_trim(skb, skb->len + diff);
+	} else if (diff <= skb_tailroom(skb)) {
+		skb_put(skb, diff);
+		memmove(o_buf + n_len, o_buf + o_len, o_left);
+		memcpy(o_buf, n_buf, n_len);
+	} else {
+		if (pskb_expand_head(skb, skb_headroom(skb), diff, pri))
+			return -ENOMEM;
+		skb_put(skb, diff);
+		memmove(skb->data + o_offset + n_len,
+			skb->data + o_offset + o_len, o_left);
+		memcpy(skb->data + o_offset, n_buf, n_len);
 	}
 
+	/* must update the iph total length here */
+	iph = skb->nh.iph;
+	iph->tot_len = htons(skb->len);
+
 	LeaveFunction(9);
-	return n_skb;
+	return 0;
 }
 
 
