@@ -5,7 +5,7 @@
  *              high-performance and highly available server based on a
  *              cluster of servers.
  *
- * Version:     $Id: ip_vs_conn.c,v 1.21 2001/10/19 15:05:17 wensong Exp $
+ * Version:     $Id: ip_vs_conn.c,v 1.22 2001/11/04 08:58:43 wensong Exp $
  *
  * Authors:     Wensong Zhang <wensong@linuxvirtualserver.org>
  *              Peter Kese <peter.kese@ijs.si>
@@ -158,7 +158,6 @@ int ip_vs_conn_hash(struct ip_vs_conn *cp)
 
 /*
  *	UNhashes ip_vs_conn from ip_vs_conn_tab.
- *	should be called with locked tables.
  *	returns bool success.
  */
 int ip_vs_conn_unhash(struct ip_vs_conn *cp)
@@ -661,19 +660,14 @@ static int ip_vs_null_xmit(struct sk_buff *skb, struct ip_vs_conn *cp)
 static int ip_vs_nat_xmit(struct sk_buff *skb, struct ip_vs_conn *cp)
 {
 	struct rtable *rt;		/* Route to the other host */
-	struct iphdr  *iph = skb->nh.iph;
+	struct iphdr  *iph;
 	union ip_vs_tphdr h;
+	int ihl;
 	unsigned short size;
 	int mtu;
 
 	EnterFunction(10);
 
-	/* do more header checking (size) */
-	h.raw = (char*) iph + (iph->ihl << 2);
-	size = ntohs(iph->tot_len) - (iph->ihl << 2);
-	if (ip_vs_header_check(iph->protocol, h.raw, size) == -1)
-		return NF_DROP;
-
 	/*
 	 * If it has ip_vs_app helper, the helper may change the payload,
 	 * so it needs full checksum checking and checksum caculation.
@@ -683,6 +677,15 @@ static int ip_vs_nat_xmit(struct sk_buff *skb, struct ip_vs_conn *cp)
 	 * do final checksum checking.
 	 */
 
+	if (cp->app && skb_is_nonlinear(skb)
+	    && skb_linearize(skb, GFP_ATOMIC) != 0)
+		return NF_DROP;
+
+	iph = skb->nh.iph;
+	ihl = iph->ihl << 2;
+	h.raw = (char*) iph + ihl;
+	size = ntohs(iph->tot_len) - ihl;
+
 	/* do TCP/UDP checksum checking if it has application helper */
 	if (cp->app && (iph->protocol != IPPROTO_UDP || h.uh->check != 0)) {
 		switch (skb->ip_summed) {
@@ -735,7 +738,7 @@ static int ip_vs_nat_xmit(struct sk_buff *skb, struct ip_vs_conn *cp)
 	skb->dst = &rt->u.dst;
 
 	/* copy-on-write the packet before mangling it */
-	if ((skb = vs_skb_cow(&skb, &iph, &h.raw)) == NULL)
+	if (ip_vs_skb_cow(&skb, &iph, &h.raw) == NULL)
 		return NF_DROP;
 
 	/* mangle the packet */
@@ -747,12 +750,10 @@ static int ip_vs_nat_xmit(struct sk_buff *skb, struct ip_vs_conn *cp)
 	 *	will fix ip_vs_conn and iph ack_seq stuff
 	 */
 	if (ip_vs_app_pkt_in(cp, &skb) != 0) {
-		/*
-		 *	skb has changed, update pointers.
-		 */
+		/* skb has probably changed, update pointers. */
 		iph = skb->nh.iph;
-		h.raw = (char*) iph + iph->ihl*4;
-		size = ntohs(iph->tot_len) - (iph->ihl * 4);
+		h.raw = (char*) iph + ihl;
+		size = skb->len - ihl;
 	}
 
 	/*
@@ -802,9 +803,9 @@ static int ip_vs_nat_xmit(struct sk_buff *skb, struct ip_vs_conn *cp)
 	LeaveFunction(10);
 	return NF_STOLEN;
 
- tx_error_icmp:
+  tx_error_icmp:
 	dst_link_failure(skb);
- tx_error:
+  tx_error:
 	kfree_skb(skb);
 	return NF_STOLEN;
 }
