@@ -5,7 +5,7 @@
  *              high-performance and highly available server based on a
  *              cluster of servers.
  *
- * Version:     $Id: ip_vs_core.c,v 1.23 2001/10/19 15:05:17 wensong Exp $
+ * Version:     $Id: ip_vs_core.c,v 1.24 2001/11/04 08:58:43 wensong Exp $
  *
  * Authors:     Wensong Zhang <wensong@linuxvirtualserver.org>
  *              Peter Kese <peter.kese@ijs.si>
@@ -332,7 +332,8 @@ ip_vs_sched_persist(struct ip_vs_service *svc, struct iphdr *iph)
  *  It selects a server according to the virtual service, and
  *  creates a connection entry.
  */
-struct ip_vs_conn *ip_vs_schedule(struct ip_vs_service *svc, struct iphdr *iph)
+static struct ip_vs_conn *
+ip_vs_schedule(struct ip_vs_service *svc, struct iphdr *iph)
 {
 	struct ip_vs_conn *cp = NULL;
 	struct ip_vs_dest *dest;
@@ -372,7 +373,8 @@ struct ip_vs_conn *ip_vs_schedule(struct ip_vs_service *svc, struct iphdr *iph)
 			    0,
 			    dest);
 	if (cp == NULL) {
-		IP_VS_ERR("Schedule: ip_vs_conn_new failed\n");
+		if (net_ratelimit())
+			IP_VS_ERR("Schedule: ip_vs_conn_new failed\n");
 		return NULL;
 	}
 
@@ -503,16 +505,34 @@ static unsigned int ip_vs_post_routing(unsigned int hooknum,
 static int ip_vs_out_icmp(struct sk_buff **skb_p)
 {
 	struct sk_buff	*skb   = *skb_p;
-	struct iphdr	*iph   = skb->nh.iph;
+	struct iphdr	*iph;
 	struct icmphdr	*icmph;
 	struct iphdr	*ciph;	/* The ip header contained within the ICMP */
 	__u16		*pptr;	/* port numbers from TCP/UDP contained header */
-	unsigned short	 len;
+	int		need_check;
+	unsigned short	ihl;
+	unsigned short	len;
 	unsigned short	clen, csize;
 	struct ip_vs_conn *cp;
 
-	icmph = (struct icmphdr *)((char *)iph + (iph->ihl<<2));
-	len   = ntohs(iph->tot_len) - (iph->ihl << 2);
+	/* reassemble IP fragments, but will it happen in ICMP packets?? */
+	need_check = 0;
+	if (skb->nh.iph->frag_off & htons(IP_MF|IP_OFFSET)) {
+		skb = ip_defrag(skb);
+		if (!skb) {
+			return NF_STOLEN;
+		}
+		*skb_p = skb;
+		need_check = 1;
+	}
+
+	if (skb_is_nonlinear(skb) && skb_linearize(skb, GFP_ATOMIC) != 0)
+		return NF_DROP;
+
+	iph = skb->nh.iph;
+	ihl = iph->ihl << 2;
+	icmph = (struct icmphdr *)((char *)iph + ihl);
+	len   = ntohs(iph->tot_len) - ihl;
 	if (len < sizeof(struct icmphdr))
 		return NF_DROP;
 
@@ -529,8 +549,11 @@ static int ip_vs_out_icmp(struct sk_buff **skb_p)
 	 */
 	if ((icmph->type != ICMP_DEST_UNREACH) &&
 	    (icmph->type != ICMP_SOURCE_QUENCH) &&
-	    (icmph->type != ICMP_TIME_EXCEEDED))
+	    (icmph->type != ICMP_TIME_EXCEEDED)) {
+		if (need_check)
+			ip_send_check(iph);
 		return NF_ACCEPT;
+	}
 
 	/* Now find the contained IP header */
 	clen = len - sizeof(struct icmphdr);
@@ -542,8 +565,11 @@ static int ip_vs_out_icmp(struct sk_buff **skb_p)
 		return NF_DROP;
 
 	/* We are only interested ICMPs generated from TCP or UDP packets */
-	if ((ciph->protocol != IPPROTO_UDP) && (ciph->protocol != IPPROTO_TCP))
+	if (ciph->protocol != IPPROTO_UDP && ciph->protocol != IPPROTO_TCP) {
+		if (need_check)
+			ip_send_check(iph);
 		return NF_ACCEPT;
+	}
 
 	/* We need at least TCP/UDP ports here */
 	if (clen < csize + sizeof(struct udphdr))
@@ -572,8 +598,11 @@ static int ip_vs_out_icmp(struct sk_buff **skb_p)
 	/* ciph content is actually <protocol, caddr, cport, daddr, dport> */
 	cp = ip_vs_conn_out_get(ciph->protocol, ciph->daddr, pptr[1],
 				ciph->saddr, pptr[0]);
-	if (cp == NULL)
+	if (!cp) {
+		if (need_check)
+			ip_send_check(iph);
 		return NF_ACCEPT;
+	}
 
 	if (IP_VS_FWD_METHOD(cp) != 0) {
 		IP_VS_ERR("shouldn't reach here, because the box is on the"
@@ -624,26 +653,19 @@ static unsigned int ip_vs_out(unsigned int hooknum,
 			      int (*okfn)(struct sk_buff *))
 {
 	struct sk_buff  *skb = *skb_p;
-	struct iphdr	*iph = skb->nh.iph;
+	struct iphdr	*iph;
 	union ip_vs_tphdr h;
 	struct ip_vs_conn *cp;
 	int size;
+	int ihl;
+	int need_check;
 
 	EnterFunction(11);
 
 	if (skb->nfcache & NFC_IPVS_PROPERTY)
 		return NF_ACCEPT;
 
-	if (iph->frag_off & htons(IP_MF|IP_OFFSET)) {
-		skb = ip_defrag(skb);
-		if (!skb) {
-			return NF_STOLEN;
-		}
-
-		iph = skb->nh.iph;
-		*skb_p = skb;
-	}
-
+	iph = skb->nh.iph;
 	if (iph->protocol == IPPROTO_ICMP)
 		return ip_vs_out_icmp(skb_p);
 
@@ -651,10 +673,30 @@ static unsigned int ip_vs_out(unsigned int hooknum,
 	if (iph->protocol != IPPROTO_TCP && iph->protocol != IPPROTO_UDP)
 		return NF_ACCEPT;
 
+	/* reassemble IP fragments */
+	need_check = 0;
+	if (iph->frag_off & htons(IP_MF|IP_OFFSET)) {
+		skb = ip_defrag(skb);
+		if (!skb) {
+			return NF_STOLEN;
+		}
+		iph = skb->nh.iph;
+		*skb_p = skb;
+		need_check = 1;
+	}
+
+	/* make sure that protocol header available in skb data area,
+	   note that skb data area may be reallocated. */
+	ihl = iph->ihl << 2;
+	if (ip_vs_header_check(skb, iph->protocol, ihl) == -1)
+		return NF_DROP;
+
+	iph = skb->nh.iph;
+	h.raw = (char*) iph + ihl;
+
 	/*
 	 *	Check if the packet belongs to an old entry
 	 */
-	h.raw = (char*) iph + (iph->ihl << 2);
 	cp = ip_vs_conn_out_get(iph->protocol, iph->saddr, h.portp[0],
 				iph->daddr, h.portp[1]);
 	if (!cp) {
@@ -676,18 +718,11 @@ static unsigned int ip_vs_out(unsigned int hooknum,
 			  ip_vs_proto_name(iph->protocol),
 			  NIPQUAD(iph->daddr),
 			  ntohs(h.portp[1]));
+		if (need_check)
+			ip_send_check(iph);
 		return NF_ACCEPT;
 	}
 
-	/* do more header checking (size) */
-	size = ntohs(iph->tot_len) - (iph->ihl << 2);
-	if (ip_vs_header_check(iph->protocol, h.raw, size) == -1) {
-		ip_vs_conn_put(cp);
-		return NF_DROP;
-	}
-	IP_VS_DBG(11, "O-pkt: %s size=%d\n",
-		  ip_vs_proto_name(iph->protocol), size);
-
 	/*
 	 * If it has ip_vs_app helper, the helper may change the payload,
 	 * so it needs full checksum checking and checksum caculation.
@@ -696,6 +731,19 @@ static unsigned int ip_vs_out(unsigned int hooknum,
 	 * do final checksum checking.
 	 */
 
+	if (cp->app && skb_is_nonlinear(skb)) {
+		if (skb_linearize(skb, GFP_ATOMIC) != 0) {
+			ip_vs_conn_put(cp);
+			return NF_DROP;
+		}
+		iph = skb->nh.iph;
+		h.raw = (char*) iph + ihl;
+	}
+
+	size = skb->len - ihl;
+	IP_VS_DBG(11, "O-pkt: %s size=%d\n",
+		  ip_vs_proto_name(iph->protocol), size);
+
 	/* do TCP/UDP checksum checking if it has application helper */
 	if (cp->app && (iph->protocol != IPPROTO_UDP || h.uh->check != 0)) {
 		switch (skb->ip_summed) {
@@ -722,24 +770,19 @@ static unsigned int ip_vs_out(unsigned int hooknum,
 		  NIPQUAD(iph->saddr), ntohs(h.portp[0]),
 		  NIPQUAD(iph->daddr), ntohs(h.portp[1]));
 
-	/*
-	 *	Set iph addr and port from ip_vs obj.
-	 */
+	/* mangle the packet */
 	iph->saddr = cp->vaddr;
 	h.portp[0] = cp->vport;
 
 	/*
-	 *	Attempt ip_vs_app call.
-	 *	will fix ip_vs and iph seq stuff
+	 *	Call application helper if needed
 	 */
 	if (ip_vs_app_pkt_out(cp, skb_p) != 0) {
-		/*
-		 *	skb has possibly changed, update pointers.
-		 */
+		/* skb has probably changed, update pointers */
 		skb = *skb_p;
 		iph = skb->nh.iph;
-		h.raw = (char*) iph + iph->ihl *4;
-		size = skb->len - (h.raw - skb->nh.raw);
+		h.raw = (char*)iph + ihl;
+		size = skb->len - ihl;
 	}
 
 	/*
@@ -836,7 +879,6 @@ static int ip_vs_in_icmp(struct sk_buff **skb_p)
 	struct rtable *rt;			/* Route to the other host */
 	int    mtu;
 
-	/* Sometimes we come here from ip_vs_forward_icmp */
 	if (skb_is_nonlinear(skb) && skb_linearize(skb, GFP_ATOMIC) != 0)
 		return NF_DROP;
 
@@ -915,17 +957,18 @@ static int ip_vs_in_icmp(struct sk_buff **skb_p)
 		return ret;
 	}
 
-	/* The ICMP packet for VS/NAT must be written to correct addresses
-	   before being forwarded to the right server */
-	if ((skb=vs_skb_cow(skb_p, &iph, (unsigned char**)&icmph)) == NULL) {
+	/* copy-on-write the packet before mangling it */
+	if ((skb = ip_vs_skb_cow(skb_p, &iph,
+				 (unsigned char**)&icmph)) == NULL) {
 		ip_vs_conn_put(cp);
 		return NF_DROP;
 	}
-
 	ciph = (struct iphdr *) (icmph + 1);
-	pptr = (__u16 *)&(((char *)ciph)[ciph->ihl*4]);
+	pptr = (__u16 *)&(((char *)ciph)[csize]);
+
+	/* The ICMP packet for VS/NAT must be written to correct addresses
+	   before being forwarded to the right server */
 
-	/* Now we do real damage to this packet...! */
 	/* First change the dest IP address, and recalc checksum */
 	iph->daddr = cp->daddr;
 	ip_send_check(iph);
@@ -997,6 +1040,8 @@ static unsigned int ip_vs_in(unsigned int hooknum,
 	union ip_vs_tphdr h;
 	struct ip_vs_conn *cp;
 	struct ip_vs_service *svc;
+	int ihl;
+	int ret;
 
 	/*
 	 *	Big tappo: only PACKET_HOST (nor loopback neither mcasts)
@@ -1017,11 +1062,13 @@ static unsigned int ip_vs_in(unsigned int hooknum,
 	if (iph->protocol != IPPROTO_TCP && iph->protocol != IPPROTO_UDP)
 		return NF_ACCEPT;
 
-	h.raw = (char*) iph + iph->ihl * 4;
-	IP_VS_DBG(11, "Incoming %s %u.%u.%u.%u:%d->%u.%u.%u.%u:%d\n",
-		  ip_vs_proto_name(iph->protocol),
-		  NIPQUAD(iph->saddr), ntohs(h.portp[0]),
-		  NIPQUAD(iph->daddr), ntohs(h.portp[1]));
+	/* make sure that protocol header available in skb data area,
+	   note that skb data area may be reallocated. */
+	ihl = iph->ihl << 2;
+	if (ip_vs_header_check(skb, iph->protocol, ihl) == -1)
+		return NF_DROP;
+	iph = skb->nh.iph;
+	h.raw = (char*) iph + ihl;
 
 	/*
 	 * Check if the packet belongs to an existing connection entry
@@ -1029,27 +1076,6 @@ static unsigned int ip_vs_in(unsigned int hooknum,
 	cp = ip_vs_conn_in_get(iph->protocol, iph->saddr, h.portp[0],
 			       iph->daddr, h.portp[1]);
 
-	/*
-	 * Check the server status
-	 */
-	if (cp && cp->dest && !(cp->dest->flags & IP_VS_DEST_F_AVAILABLE)) {
-		/* the destination server is not availabe */
-
-		if (sysctl_ip_vs_expire_nodest_conn) {
-			/* expire the connection immediately */
-			__ip_vs_conn_put(cp);
-			if (del_sltimer(&cp->timer))
-				ip_vs_conn_expire_now(cp);
-
-			return NF_DROP;
-		} else {
-			/* don't restart its timer, and silently
-			   drop the packet. */
-			__ip_vs_conn_put(cp);
-			return NF_DROP;
-		}
-	}
-
 	if (!cp
 	    && (h.th->syn || (iph->protocol!=IPPROTO_TCP))
 	    && (svc = ip_vs_service_get(skb->nfmark, iph->protocol,
@@ -1074,40 +1100,62 @@ static unsigned int ip_vs_in(unsigned int hooknum,
 		ip_vs_service_put(svc);
 	}
 
-	if (cp != NULL) {
-		int ret;
-
-		ip_vs_in_stats(cp, skb);
-		ip_vs_set_state(cp, VS_STATE_INPUT, iph, h.portp);
-		if (cp->packet_xmit)
-			ret = cp->packet_xmit(skb, cp);
-		else {
-			IP_VS_DBG(0, "warning: packet_xmit is null");
-			ret = NF_ACCEPT;
-		}
-
-		/* increase its packet counter and check if it is needed
-		   to be synchronized */
-		atomic_inc(&cp->in_pkts);
-		if (ip_vs_sync_state == IP_VS_STATE_MASTER
-		    && (cp->protocol != IPPROTO_TCP
-			|| cp->state == IP_VS_S_ESTABLISHED)
-		    && (atomic_read(&cp->in_pkts) % 50
-			== sysctl_ip_vs_sync_threshold))
-			ip_vs_sync_conn(cp);
-
-		ip_vs_conn_put(cp);
-		return ret;
+	if (!cp) {
+		/* sorry, all this trouble for a no-hit :) */
+		IP_VS_DBG(12, "packet for %s %d.%d.%d.%d:%d continue "
+			  "traversal as normal.\n",
+			  ip_vs_proto_name(iph->protocol),
+			  NIPQUAD(iph->daddr),
+			  ntohs(h.portp[1]));
+		return NF_ACCEPT;
 	}
 
-	/* sorry, all this trouble for a no-hit :) */
-	IP_VS_DBG(12, "packet for %s %d.%d.%d.%d:%d continue traversal "
-		  "as normal.\n",
+	IP_VS_DBG(11, "Incoming %s %u.%u.%u.%u:%d->%u.%u.%u.%u:%d\n",
 		  ip_vs_proto_name(iph->protocol),
-		  NIPQUAD(iph->daddr),
-		  ntohs(h.portp[1]));
+		  NIPQUAD(iph->saddr), ntohs(h.portp[0]),
+		  NIPQUAD(iph->daddr), ntohs(h.portp[1]));
 
-	return NF_ACCEPT;
+	/*
+	 * Check the server status
+	 */
+	if (cp->dest && !(cp->dest->flags & IP_VS_DEST_F_AVAILABLE)) {
+		/* the destination server is not availabe */
+
+		if (sysctl_ip_vs_expire_nodest_conn) {
+			/* expire the connection immediately */
+			__ip_vs_conn_put(cp);
+			if (del_sltimer(&cp->timer))
+				ip_vs_conn_expire_now(cp);
+
+			return NF_DROP;
+		} else {
+			/* don't restart its timer, and silently
+			   drop the packet. */
+			__ip_vs_conn_put(cp);
+			return NF_DROP;
+		}
+	}
+
+	ip_vs_in_stats(cp, skb);
+	ip_vs_set_state(cp, VS_STATE_INPUT, iph, h.portp);
+	if (cp->packet_xmit)
+		ret = cp->packet_xmit(skb, cp);
+	else {
+		IP_VS_DBG(0, "warning: packet_xmit is null");
+		ret = NF_ACCEPT;
+	}
+
+	/* increase its packet counter and check if it is needed
+	   to be synchronized */
+	atomic_inc(&cp->in_pkts);
+	if (ip_vs_sync_state == IP_VS_STATE_MASTER
+	    && (cp->protocol != IPPROTO_TCP
+		|| cp->state == IP_VS_S_ESTABLISHED)
+	    && (atomic_read(&cp->in_pkts) % 50 == sysctl_ip_vs_sync_threshold))
+		ip_vs_sync_conn(cp);
+
+	ip_vs_conn_put(cp);
+	return ret;
 }
 
 
@@ -1126,12 +1174,26 @@ static unsigned int ip_vs_forward_icmp(unsigned int hooknum,
 				       const struct net_device *out,
 				       int (*okfn)(struct sk_buff *))
 {
-	struct iphdr	*iph   = (*skb_p)->nh.iph;
+	struct sk_buff	*skb = *skb_p;
+	struct iphdr	*iph = skb->nh.iph;
 
-	if (iph->protocol == IPPROTO_ICMP)
-		return ip_vs_in_icmp(skb_p);
+	if (iph->protocol != IPPROTO_ICMP)
+		return NF_ACCEPT;
 
-	return NF_ACCEPT;
+	if (iph->frag_off & htons(IP_MF|IP_OFFSET)) {
+		int ret;
+		
+		skb = ip_defrag(skb);
+		if (!skb)
+			return NF_STOLEN;
+		*skb_p = skb;
+		ret = ip_vs_in_icmp(skb_p);
+		if (ret == NF_ACCEPT)
+			ip_send_check((*skb_p)->nh.iph);
+		return ret;
+	}
+
+	return ip_vs_in_icmp(skb_p);
 }
 
 
