@@ -1,7 +1,7 @@
 /*
  * IP_VS        ftp application module
  *
- * Version:	$Id: ip_vs_ftp.c,v 1.8 2001/10/19 15:05:17 wensong Exp $
+ * Version:	$Id: ip_vs_ftp.c,v 1.9 2001/11/04 08:58:43 wensong Exp $
  *
  * Authors:	Wensong Zhang <wensong@linuxvirtualserver.org>
  *
@@ -161,7 +161,7 @@ static int ip_vs_ftp_out(struct ip_vs_app *mapp,
 		iph = skb->nh.iph;
 		th = (struct tcphdr *)&(((char *)iph)[iph->ihl*4]);
 		data = (char *)th + (th->doff << 2);
-		data_limit = skb->h.raw + skb->len;
+		data_limit = skb->tail;
 
 		if (ip_vs_ftp_get_addrport(data, data_limit,
 					   SERVER_STRING,
@@ -206,30 +206,23 @@ static int ip_vs_ftp_out(struct ip_vs_app *mapp,
 			port&255, port>>8&255);
 		buf_len = strlen(buf);
 
-		IP_VS_DBG(1-debug, "new PORT %u.%u.%u.%u:%d\n",
-			  NIPQUAD(from), ntohs(port));
-
 		/*
 		 * Calculate required delta-offset to keep TCP happy
 		 */
 		diff = buf_len - (end-start);
 
-		/*
-		 *	No shift.
-		 */
-		if (diff==0) {
-			/*
-			 * simple case, just replace the old PORT cmd
-			 */
+		if (diff == 0) {
+			/* simply replace it with new passive address */
 			memcpy(start, buf, buf_len);
 		} else {
-			*skb_p = ip_vs_skb_replace(skb, GFP_ATOMIC, start,
-						   end-start, buf, buf_len);
+			/* fixme: return value isn't checked here */
+			ip_vs_skb_replace(skb, GFP_ATOMIC, start,
+					  end-start, buf, buf_len);
 		}
 
 		cp->app_data = NULL;
+		ip_vs_conn_listen(n_cp);
 		ip_vs_conn_put(n_cp);
-
 		return diff;
 	}
 	return 0;
@@ -274,7 +267,7 @@ static int ip_vs_ftp_in(struct ip_vs_app *mapp,
 	   the length of the header in 32-bit multiples, it is accurate
 	   to caculate data address by th+HLEN*4 */
 	data = data_start = (char *)th + (th->doff << 2);
-	data_limit = skb->h.raw + skb->len;
+	data_limit = skb->tail;
 
 	while (data < data_limit) {
 		if (strnicmp(data, "PASV\r\n", 6) == 0) {
