@@ -2,7 +2,7 @@
  *      ipvsadm - IP Virtual Server ADMinistration program
  *                for IPVS NetFilter Module in kernel 2.4
  *
- *      Version: $Id: ipvsadm.c,v 1.15 2001/10/19 15:05:17 wensong Exp $
+ *      Version: $Id: ipvsadm.c,v 1.17 2001/11/04 09:18:01 wensong Exp $
  *
  *      Authors: Wensong Zhang <wensong@linuxvirtualserver.org>
  *               Peter Kese <peter.kese@ijs.si>
@@ -125,12 +125,12 @@
 #include "libipvs/libipvs.h"
 
 #define IPVSADM_VERSION_NO              "v" VERSION
-#define IPVSADM_VERSION_DATE            "2001/09/18"
+#define IPVSADM_VERSION_DATE            "2001/11/04"
 #define IPVSADM_VERSION         IPVSADM_VERSION_NO " " IPVSADM_VERSION_DATE
 
 #define MINIMUM_IPVS_VERSION_MAJOR      0
 #define MINIMUM_IPVS_VERSION_MINOR      9
-#define MINIMUM_IPVS_VERSION_PATCH      4
+#define MINIMUM_IPVS_VERSION_PATCH      6
 
 #ifndef IPVS_VERSION
 #define IPVS_VERSION(x,y,z) (((x)<<16)+((y)<<8)+(z))
@@ -155,6 +155,7 @@ static const char* cmdnames[] = {
 	"stop-daemon",
 	"restore",
 	"save",
+	"zero",
 };
 
 #define OPT_NONE	0x00000
@@ -216,6 +217,7 @@ static const char commands_v_options[NUMBER_OF_CMD][NUMBER_OF_OPT] =
 /*STOP-D*/    {'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
 /*RESTORE*/   {'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
 /*SAVE*/      {' ', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
+/*ZERO*/      {'x', 'x', ' ', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x'},
 };
 
 /* printing format flags */
@@ -324,6 +326,7 @@ int parse_options(int argc, char **argv, struct ip_vs_rule_user *ur,
 		{"clear", 'C', POPT_ARG_NONE, NULL, 'C'},
 		{"list", 'L', POPT_ARG_NONE, NULL, 'L'},
 		{"list", 'l', POPT_ARG_NONE, NULL, 'l'},
+		{"zero", 'Z', POPT_ARG_NONE, NULL, 'Z'},
 		{"add-server", 'a', POPT_ARG_NONE, NULL, 'a'},
 		{"edit-server", 'e', POPT_ARG_NONE, NULL, 'e'},
 		{"delete-server", 'd', POPT_ARG_NONE, NULL, 'd'},
@@ -388,6 +391,9 @@ int parse_options(int argc, char **argv, struct ip_vs_rule_user *ur,
 	case 'l':
 		set_command(cmd, IP_VS_SO_SET_LIST);
 		break;
+	case 'Z':
+		set_command(cmd, IP_VS_SO_SET_ZERO);
+		break;
 	case '4':
 		set_command(cmd, IP_VS_SO_SET_TIMEOUTS);
 		break;
@@ -557,7 +563,7 @@ int parse_options(int argc, char **argv, struct ip_vs_rule_user *ur,
 		  int *cmd, unsigned int *options, unsigned int *format)
 {
 	int c, parse;
-	const char *optstring = "AEDCSRaedlLhvt:u:f:s:M:p::w:r:gmicn";
+	const char *optstring = "AEDCZSRaedlLhvt:u:f:s:M:p::w:r:gmicn";
 	const struct option long_options[] =
 	{
 		{"add-service", 0, 0, 'A'},
@@ -565,6 +571,7 @@ int parse_options(int argc, char **argv, struct ip_vs_rule_user *ur,
 		{"delete-service", 0, 0, 'D'},
 		{"clear", 0, 0, 'C'},
 		{"list", 0, 0, 'L'},
+		{"zero", 0, 0, 'Z'},
 		{"add-server", 0, 0, 'a'},
 		{"edit-server", 0, 0, 'e'},
 		{"delete-server", 0, 0, 'd'},
@@ -602,7 +609,6 @@ int parse_options(int argc, char **argv, struct ip_vs_rule_user *ur,
 	extern int opterr;
 	extern int optopt;
 
-	/* Re-process the arguments each time options is called */
 	optind = 1;
 
 	if ((c = getopt_long(argc, argv, optstring,
@@ -635,6 +641,9 @@ int parse_options(int argc, char **argv, struct ip_vs_rule_user *ur,
 	case 'l':
 		set_command(cmd, IP_VS_SO_SET_LIST);
 		break;
+	case 'Z':
+		set_command(cmd, IP_VS_SO_SET_ZERO);
+		break;
 	case '4':
 		set_command(cmd, IP_VS_SO_SET_TIMEOUTS);
 		if (optind + 1 < argc
@@ -1105,11 +1114,12 @@ void usage_exit(const char *program, const int exit_status)
 		"  %s -a|e -t|u|f service-address -r server-address [-g|i|m] [-w weight]\n"
 		"  %s -d -t|u|f service-address -r server-address\n"
 		"  %s -L|l [options]\n"
+		"  %s -Z [-t|u|f service-address]\n"
 		"  %s --set tcp tcpfin udp\n"
 		"  %s --start-daemon state [--mcast-interface interface]\n"
 		"  %s --stop-daemon\n"
 		"  %s -h\n\n",
-		program, program,
+		program, program, program,
 		program, program, program, program, program,
 		program, program, program, program, program);
 
@@ -1126,6 +1136,7 @@ void usage_exit(const char *program, const int exit_status)
 		"  --edit-server     -e        edit real server with options\n"
 		"  --delete-server   -d        delete real server\n"
 		"  --list            -L|-l     list the table\n"
+		"  --zero            -Z        zero counters in a service or all services\n"
 		"  --set tcp tcpfin udp        set connection timeout values\n"
 		"  --start-daemon              start connection sync daemon\n"
 		"  --stop-daemon               stop connection sync daemon\n"
@@ -1517,8 +1528,10 @@ void list_service(struct ip_vs_rule_user *ur, unsigned int format)
 	struct ip_vs_service_user *svc;
 
 	if (!(svc = ipvs_get_service(ur->vfwmark, ur->protocol,
-				     ur->vaddr, ur->vport)))
+				     ur->vaddr, ur->vport))) {
+		fprintf(stderr, "%s\n", ipvs_strerror(errno));
 		exit(1);
+	}
 
 	print_title(format);
 	print_service(svc, format);
@@ -1535,8 +1548,10 @@ void list_all(unsigned int format)
 		printf("IP Virtual Server version %d.%d.%d (size=%d)\n",
 		       NVERSION(ipvs_info.version), ipvs_info.size);
 
-	if (!(get = ipvs_get_services()))
+	if (!(get = ipvs_get_services())) {
+		fprintf(stderr, "%s\n", ipvs_strerror(errno));
 		exit(1);
+	}
 
 	print_title(format);
 	for (i=0; i<get->num_services; i++)
