@@ -1,7 +1,7 @@
 /*
  * libipvs:	Library for manipulating IPVS through [gs]etsockopt
  *
- * Version:     $Id: libipvs.c,v 1.2 2001/10/19 15:05:17 wensong Exp $
+ * Version:     $Id: libipvs.c,v 1.3 2001/11/04 08:59:01 wensong Exp $
  *
  * Authors:     Wensong Zhang <wensong@linuxvirtualserver.org>
  *
@@ -77,6 +77,7 @@ struct ip_vs_get_services *ipvs_get_services(void)
 	if (!(get = malloc(len)))
 		return NULL;
 
+	ipvs_cmd = IP_VS_SO_GET_SERVICES;
 	get->num_services = ipvs_info.num_services;
 	if (getsockopt(sockfd, IPPROTO_IP,
 		       IP_VS_SO_GET_SERVICES, get, &len) < 0) {
@@ -96,6 +97,7 @@ struct ip_vs_get_dests *ipvs_get_dests(struct ip_vs_service_user *svc)
 	if (!(d = malloc(len)))
 		return NULL;
 
+	ipvs_cmd = IP_VS_SO_GET_DESTS;
 	d->fwmark = svc->fwmark;
 	d->protocol = svc->protocol;
 	d->addr = svc->addr;
@@ -120,6 +122,7 @@ ipvs_get_service(__u32 fwmark, __u16 protocol, __u32 vaddr, __u16 vport)
 	if (!(svc = malloc(len)))
 		return NULL;
 
+	ipvs_cmd = IP_VS_SO_GET_SERVICE;
 	svc->fwmark = fwmark;
 	svc->protocol = protocol;
 	svc->addr = vaddr;
@@ -142,6 +145,7 @@ struct ip_vs_timeout_user *ipvs_get_timeouts(void)
 	if (!(u = malloc(len)))
 		return NULL;
 
+	ipvs_cmd = IP_VS_SO_GET_TIMEOUTS;
 	if (getsockopt(sockfd, IPPROTO_IP, IP_VS_SO_GET_TIMEOUTS,
 		       (char *)u, &len)) {
 		free(u);
@@ -160,6 +164,7 @@ struct ip_vs_daemon_user *ipvs_get_daemon(void)
 	if (!(u = malloc(len)))
 		return NULL;
 
+	ipvs_cmd = IP_VS_SO_GET_DAEMON;
 	if (getsockopt(sockfd, IPPROTO_IP,
 		       IP_VS_SO_GET_DAEMON, (char *)u, &len)) {
 		free(u);
@@ -198,6 +203,11 @@ const char *ipvs_strerror(int err)
 	    { IP_VS_SO_SET_EDITDEST, ENOENT, "No such destination" },
 	    { IP_VS_SO_SET_DELDEST, ESRCH, "Service not defined" },
 	    { IP_VS_SO_SET_DELDEST, ENOENT, "No such destination" },
+	    { IP_VS_SO_SET_STARTDAEMON, EEXIST, "Daemon has already run" },
+	    { IP_VS_SO_SET_STOPDAEMON, ESRCH, "No daemon is running" },
+	    { IP_VS_SO_SET_STOPDAEMON, ESRCH, "No daemon is running" },
+	    { IP_VS_SO_GET_SERVICE, ESRCH, "No such service" },
+	    { IP_VS_SO_GET_DESTS, ESRCH, "No such service" },
 	  };
 
 	for (i = 0; i < sizeof(table)/sizeof(struct table_struct); i++) {
